package TMApplet;

import java.applet.*;
import java.awt.*;
import java.util.*;

/**
 * Maintain the state transition table.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: States.java,v 1.1.1.1 2002/02/28 12:09:45 peteg Exp $
 */
class States extends Panel {
    private static class State {
	int tape;
	int newSymbol;
	char direction;
	int newState;
    }

    private TMApplet parent;
    private Vector stateTable = new Vector();
    private int currentState = 0;
    private int numStates = 0;
    private int numTransitions = 0;
    private boolean halted = false;

    // Symbol table
    //  hash maps names -> numbers, vector maps numbers -> names
    //  bitSet tracks breakpoint states
    private Hashtable symbolNames = new Hashtable();
    private Vector stateNumbers = new Vector();
    private BitSet breakPoints = new BitSet();

    // Off-screen graphics buffer
    private Dimension offDim;
    private Image offImage;
    private Graphics offGraphics;

    public Insets getInsets() {
	return new Insets(5,5,5,5);
    }

    public Dimension getMinimumSize() {
	return getPreferredSize();
    }

    public Dimension getPreferredSize() {
	return new Dimension(parent.getSize().width / 2, 100);
    }

    States(TMApplet tmApplet) {
	parent = tmApplet;
    }

    public void reset() {
	currentState = 0;
	numTransitions = 0;
	halted = false;
	update(this.getGraphics());
    }

    // Inspector operations
    public boolean Halted() {
        return halted;
    }

    public int CurrentState() {
        return currentState;
    }

    public int NumTransitions() {
        return numTransitions;
    }

    // Map symbolic names -> state numbers
    private int name2num(String name) {
	Integer num;

	if((num = (Integer)symbolNames.get((Object)name)) != null)
	    // Existing state
	    return num.intValue();
	else {
	    // New state
	    symbolNames.put((Object)name, (Object)(new Integer(numStates)));
	    stateNumbers.addElement((Object)name);
	    return numStates++;
	}
    }

    // Map state numbers -> symbolic names
    private String num2name(int num) {
	if(num > stateNumbers.size()) {
	    return "Halt";
	} else {
	    return (String)stateNumbers.elementAt(num);
	}
    }
    // Insert states into the state table
    private void insertState(int state, int symbol, int tape, State newState) {
        // Need to make sure vector is long enough
        while(stateTable.size() <= state)
	    stateTable.addElement((Object)new State[parent.alphabetLength()][parent.tapeHeads.length]);
        ((State[][])stateTable.elementAt(state))[symbol][tape] = newState;
    }

    // Lookup the state table
    private State lookupState(int state, int symbol, int tape) {
        if(state > (stateTable.size() - 1))
            return null;
        else
            return ((State[][])stateTable.elementAt(state))[symbol][tape];
    }

    public void newState(String s) {
	State newState = new State();
	StringTokenizer st = new StringTokenizer(s, ",", false);
	int initState, initSymbol, tape;

	try {
	    initState = name2num(st.nextToken());
	    tape = (parent.tapeHeads.length > 1) ? (Integer.parseInt(st.nextToken()) - 1) : 0;
	    initSymbol = parent.sym2int(st.nextToken().charAt(0));
	} catch(Exception e) {
	    parent.dPrint("STATE: malformed state tuple.");
	    return;
	}

	try {
	    newState.tape = (parent.tapeHeads.length > 1) ? (Integer.parseInt(st.nextToken()) - 1) : 0;
	    newState.newSymbol = parent.sym2int(st.nextToken().charAt(0));
	    newState.direction = Character.toUpperCase(st.nextToken().charAt(0));
	    newState.newState = name2num(st.nextToken());
	} catch(Exception e) {
	    parent.dPrint("STATE: ignoring: state: " + initState + " symbol: '" + parent.int2sym(initSymbol) + "'");
	    return;
	}

	// Sanity checks
	if(tape >= parent.tapeHeads.length || newState.tape >= parent.tapeHeads.length) {
	    parent.dPrint("STATE: bad tape index: " + tape + " or " + newState.tape +
			  " ignoring: state: " + initState + " tape symbol: '" + parent.int2sym(initSymbol) + "'");
	    return;
	}

        // Add to state table
	insertState(initState, initSymbol, tape, newState);
    }

    // Parse breakpoint info
    public void parseBP(String s) {
	StringTokenizer st = new StringTokenizer(s, ",", false);

	while(st.hasMoreTokens()) {
	    try {
		breakPoints.set(name2num(st.nextToken()));
	    } catch(Exception e) {
		parent.dPrint("STATE: malformed breakpoint.");
	    }
	}
    }

    // Check for a breakpoint on the current state
    public boolean chkBP() {
	return breakPoints.get(currentState);
    }

    // Perform a transition, update the tape
    public boolean transition() {
	int currentSymbol[] = new int[parent.tapeHeads.length];
	int tape;
	State state = null;

	if(halted) {
	    return false;
	}

	// Get the current symbol on all tapes
	for(tape = 0; tape < parent.tapeHeads.length; tape++)
	    currentSymbol[tape] = parent.tapeHeads[tape].read();

	// Get the first transition that matches a (state, symbol) pair
	for(tape = 0; tape < parent.tapeHeads.length; tape++)
	    if((state = lookupState(currentState, currentSymbol[tape], tape)) != null)
		break;

	if(state == null) {
	    // Non-existent transition = halt
            parent.dPrint("STATE: non-existent transition (halting):");
	    parent.dPrint(" state: " + num2name(currentState));
	    for(int i = 0; i < parent.tapeHeads.length; i++)
		parent.dPrint("  tape" + (i+1) + " symbol: '" + parent.int2sym(currentSymbol[i]) + "'");
	    halted = true;
	} else {
	    numTransitions++;
	    parent.tapeHeads[state.tape].write(state.newSymbol);
	    switch(state.direction) {
	    case 'L':
		parent.tapeHeads[state.tape].moveLeft();
		break;
	    case 'R':
		parent.tapeHeads[state.tape].moveRight();
		break;
	    case 'N':
		parent.tapeHeads[state.tape].drawTape();
		parent.tapeHeads[state.tape].update(parent.tapeHeads[state.tape].getGraphics());
		break;
	    default:
		parent.dPrint("STATE: unknown direction: state: " + num2name(currentState));
		break;
	    }

	    currentState = state.newState;
	}
	this.update(this.getGraphics());
	return !halted;
    }

    public void paint(Graphics g) {
	update(g);
    }

    public void update(Graphics g) {
	Dimension d = getSize();

	// Create the offscreen graphics context, if necessary
	if((offGraphics == null)
	   || (d.width != offDim.width)
	   || (d.height != offDim.height)) {

	    if(offGraphics != null)
		offGraphics.dispose();

	    offDim = d;
	    offImage = createImage(d.width, d.height);
	    offGraphics = offImage.getGraphics();
	}

	int x = 5, y = 20;
	int fontHeight = offGraphics.getFontMetrics().getHeight();
	String s;

	offGraphics.setColor(parent.bg);
	offGraphics.fillRect(0, 0, offDim.width, offDim.height);
	offGraphics.setColor(Color.black);

	for(int i = 0; i < parent.tapeHeads.length; y += fontHeight, i++)
	    offGraphics.drawString("# of non-blank symbols on tape #" + (i + 1) + ": " + parent.tapeHeads[i].NumNonBlanks(), x, y);

	offGraphics.drawString("# of transitions: " + numTransitions, x, y);
	y += fontHeight;

	if(halted) {
	    s = "Halted";
	} else {
	    s = num2name(currentState);
	    if(breakPoints.get(currentState)) {
		s += (" (break point)");
	    }
	}
	offGraphics.drawString("Current state: " + s, x, y);

	parent.lowered3DBorder(offGraphics, offDim);
	g.drawImage(offImage, 0, 0, this);
    }
}
