package TMApplet;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/*
 * Reads / writes data to / from a tape, maintaining state.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@unsw.edu.au">peteg@unsw.edu.au</A>)
 * @version $Id: TapeHead.java,v 1.1.1.1 2002/02/28 12:09:45 peteg Exp $
 */
class TapeHead extends Canvas {
    // Internal Tape state
    private Vector tneg = new Vector(), tpos = new Vector();
    private int numNonBlanks = 0;

    private Vector tnegOrig, tposOrig;
    private int numNonBlanksOrig;

    // Internal TapeHead state
    private TMApplet parent;
    private int headPos;
    private int headPosOrig;

    // Off-screen graphics buffers
    private Dimension tapeDim, offDim;
    private Image tapeImage, offImage;
    private Graphics tapeGraphics, offGraphics;

    // Graphics
    private static int cellWidth = 30, cellHeight = 30;
    private int xStart, yStart;
    private int animOffset = 0, lastX, scrollOffset = 0, centreCell, leftOffset;

    private static int maxIndexHeight = (int)(0.4 * cellHeight), maxIndexWidth = (int)(0.4 * cellWidth);
    private static int maxDataHeight = (int)(0.8 * cellHeight), maxDataWidth = (int)(0.8 * cellWidth);
    private int indexHeight, indexWidth, dataHeight, dataWidth;
    private Font dataFont, indexFont;

    public Insets getInsets() {
	return new Insets(5,5,5,5);
    }

    public Dimension getMinimumSize() {
	return getPreferredSize();
    }

    public Dimension getPreferredSize() {
	return new Dimension(parent.getSize().width, 2 * cellHeight);
    }

    TapeHead(TMApplet tmApplet, String tapeString, int tapeOffset, int tapeHeadOffset) {
	parent = tmApplet;

	if(tapeString != null)
	    for(int i = 0; i < tapeString.length(); i++)
		tapeWrite(tapeOffset + i, parent.sym2int(tapeString.charAt(i)));
	centreCell = headPosOrig = headPos = tapeHeadOffset;

	// Save for a later reset
	tnegOrig = (Vector)tneg.clone();
	tposOrig = (Vector)tpos.clone();
	numNonBlanksOrig = numNonBlanks;

	// If the machine is stopped, the user can:
	//  1. drag the tape left or right
	//  2. centre the tape head with a double click
	this.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		if(parent.thread == null && e.getClickCount() == 2) {
		    scrollOffset = 0;
		    centreCell = headPos;
		    refresh();
		}
	    }

	    public void mousePressed(MouseEvent e) {
		lastX = e.getX();
	    }
	});

	this.addMouseMotionListener(new MouseMotionAdapter() {
	    public void mouseDragged(MouseEvent e) {
		if(parent.thread == null) {
		    scrollOffset += (e.getX() - lastX);
		    centreCell -= (int)(scrollOffset / cellWidth);
		    scrollOffset %= cellWidth;
		    lastX = e.getX();
		    refresh();
		}
	    }
	});
    }

    public void reset() {
	tneg = (Vector)tnegOrig.clone();
	tpos = (Vector)tposOrig.clone();
	numNonBlanks = numNonBlanksOrig;

	// Don't move the tapehead
	centreCell = headPosOrig - (headPos - centreCell);
	headPos = headPosOrig;
	refresh();
    }

    // Read the data at the specified position
    public int tapeRead(int pos) {
	if(pos >= 0)
	    if(pos > (tpos.size() - 1))
		return TMApplet.Blank;
	    else
		return ((Integer)tpos.elementAt(pos)).intValue();
	else {
	    pos = -pos;
	    if(pos > (tneg.size() - 1))
		return TMApplet.Blank;
	    else
		return ((Integer)tneg.elementAt(pos)).intValue();
	}
    }

    // Write to the tape
    public void tapeWrite(int pos, int elt) {
	if(pos >= 0) {
	    while(tpos.size() <= pos)
		tpos.addElement(new Integer(TMApplet.Blank));

	    if(elt == TMApplet.Blank)
		numNonBlanks--;
	    if(((Integer)tpos.elementAt(pos)).intValue() == TMApplet.Blank)
		numNonBlanks++;

	    tpos.setElementAt(new Integer(elt), pos);
	} else {
	    pos = -pos;
	    while(tneg.size() <= pos)
		tneg.addElement(new Integer(TMApplet.Blank));

	    if(elt == TMApplet.Blank)
		numNonBlanks--;
	    if(((Integer)tneg.elementAt(pos)).intValue() == TMApplet.Blank)
		numNonBlanks++;

	    tneg.setElementAt(new Integer(elt), Math.abs(pos));
	}
    }

    public void moveLeft() {
	headPos--;
	animateTape(true);
    }

    public void moveRight() {
	headPos++;
	animateTape(false);
    }

    public int read() {
	return tapeRead(headPos);
    }

    public void write(int val) {
	tapeWrite(headPos, val);
    }

    public int NumNonBlanks() {
        return numNonBlanks;
    }

    // Update graphics to reflect current state of this tape
    void refresh() {
	drawTape();
	update(this.getGraphics());
    }

    private void animateTape(boolean left) {
	if(parent.controlPanel.speed() != 4) {
	    // Animate the tape movement
	    drawTape();
	    centreCell += left ? -1 : 1;
	    update(this.getGraphics());
	    while(Math.abs(animOffset) < cellWidth) {
		long startTime = System.currentTimeMillis();
		long delayTime = 0;

		animOffset = left ? animOffset + 1 : animOffset - 1;
		update(this.getGraphics());

		switch(parent.controlPanel.speed()) {
		case 1: delayTime = 500; break;
		case 2: delayTime = 250; break;
		}

		try {
		    Thread.sleep(Math.max(0, delayTime / cellWidth - (System.currentTimeMillis() - startTime)));
		} catch (InterruptedException e) { break; }
	    }
	    drawTape();
	} else {
	    centreCell += left ? -1 : 1;
	}
    }

    // Draw the tape
    // We pre-draw a fair bit of the tape so that scrolling is quicker
    void drawTape() {
	tapeGraphics.setColor(parent.bg);
	tapeGraphics.fillRect(0, 0, tapeDim.width, tapeDim.height);

        // Draw a series of squares, their numbers and contents
	int rec = 0;
	int firstCell = centreCell - leftOffset;
	StringBuffer s = new StringBuffer(" ");

	tapeGraphics.setColor(Color.red);
        for(int i = 0; rec < tapeDim.width; i++) {
	    rec = i * cellWidth;
	    tapeGraphics.drawRect(rec, 0, cellWidth, cellHeight - 1);

	    tapeGraphics.setFont(indexFont);
	    tapeGraphics.drawString(Integer.toString(firstCell + i), rec + 2, indexHeight);

	    tapeGraphics.setFont(dataFont);
	    s.setCharAt(0, parent.int2sym(tapeRead(firstCell + i)));
	    tapeGraphics.drawString(s.toString(), rec + (cellWidth + dataWidth) / 2, (cellHeight + dataHeight) / 2);
        }
	animOffset = 0;
    }
	
    public void paint(Graphics g) {
	update(g);
    }

    public void update(Graphics g) {
	Dimension d = getSize();

	// Create the offscreen graphics context, if necessary
	if((offGraphics == null)
	   || (offDim.width != d.width)
	   || (offDim.height != d.height)) {

	    if(tapeGraphics != null)
		tapeGraphics.dispose();
	    if(offGraphics != null)
		offGraphics.dispose();

	    // Tape off-screen buffer (for speed)
	    tapeImage = createImage(d.width + (4 * cellWidth), cellHeight);
	    tapeDim = new Dimension(d.width + (4 * cellWidth), cellHeight);
	    tapeGraphics = tapeImage.getGraphics();

	    // Usual off-screen buffer (for flicker-removal)
	    offImage = createImage(d.width, d.height);
	    offDim = d;
	    offGraphics = offImage.getGraphics();

	    // Calc where to start drawing
	    leftOffset = d.width / (2 * cellWidth) + 3;
	    xStart = (int)(((float)d.width / 2) - ((leftOffset + 0.5) * cellWidth));
	    yStart = (d.height - cellHeight) / 2;

	    // Re-size fonts so that everything fits
	    FontMetrics fontMetrics = g.getFontMetrics();

	    dataFont = g.getFont();
	    while(true) {
		if((fontMetrics.getHeight() <= maxDataHeight) && (fontMetrics.getMaxAdvance() <= maxDataWidth))
		    break;
		else {
		    g.setFont(dataFont = new Font(dataFont.getName(), dataFont.getStyle(), dataFont.getSize() - 1));
		    fontMetrics = g.getFontMetrics();
		}
	    }
	    dataHeight = fontMetrics.getHeight();
	    dataWidth = fontMetrics.getMaxAdvance();

	    indexFont = g.getFont();
	    while(true) {
		if((fontMetrics.getHeight() <= maxIndexHeight) && (fontMetrics.getMaxAdvance() <= maxIndexHeight))
		    break;
		else {
		    g.setFont(indexFont = new Font(indexFont.getName(), indexFont.getStyle(), indexFont.getSize() - 1));
		    fontMetrics = g.getFontMetrics();
		}
	    }
	    indexHeight = fontMetrics.getHeight();
	    indexWidth = fontMetrics.getMaxAdvance();

	    drawTape();
	}
	offGraphics.setColor(parent.bg);
	offGraphics.fillRect(0, 0, offDim.width, offDim.height);

	// Tape, at current animation position
	offGraphics.drawImage(tapeImage, xStart + animOffset + scrollOffset, yStart, this);

	// Hilight the headPos square
	offGraphics.setColor(Color.black);
	offGraphics.drawRect(scrollOffset + (offDim.width - cellWidth + 1) / 2 + (headPos - centreCell) * cellWidth, yStart, cellWidth, cellHeight - 1);
	offGraphics.drawRect(scrollOffset + (offDim.width - cellWidth + 1) / 2 + (headPos - centreCell) * cellWidth - 4 , yStart - 4, cellWidth + 8, cellHeight + 8);

	parent.lowered3DBorder(offGraphics, offDim);
	g.drawImage(offImage, 0, 0, this);
    }
}
